"""
Helper tools for IP related information.

**Features**:

    Uses **ipinfo.io** (https://ipinfo.io) API to retrive IP related information.

        - A free **API token** is required to call the API.
        - Tokens can be aquired by going to https://ipinfo.io/signup.
        - See below (Setting up user Token) for integrating token with this package.

    To minimize API calls (and response time):

        - IP Data is cached locally in ~/.IpHelper/cache.json.
        - MAC Data is cached locally in ~/.IpHelper/mac_info.json.

**Setting up a User Token**:

    - Go to https://ipinfo.io/signup
    - Fill out form as requested
    - Capture your token key for future reference 
    - Save your token using set-iphelper-token utility (dt-cli-tools) or manually.
    
    **Saving token manually**

        - Open an editor and load/create: ~/.IpHelper/ip_token_info.json.
        - Create one line in the file as: {"token": "xxxxxxxxxxxxxx"}
            - Replace the x's with the token supplied by ipinfo.io.
        - Save the file.



**Note**:

    For devices that are only identified by their IP and MAC address (ie. no hostname identified),
    You may assign a hostname by manually updating ~/.IpHelper/mac_info.json, 
    which is keyed by mac address.

    Format::

        {
            "XX:XX:XX:XX:XX:XX": {
                "vendor": "Ring LLC",
                "hostname": "Ring.Doorbell"
            },
        }

    One or both of the keys (vendor/hostname) must be supplied.

"""

import json
import pathlib
from datetime import datetime, timedelta
from threading import Semaphore
from typing import Dict, List, Tuple

import requests
import urllib3
from dateutil import parser
from dt_tools.logger.logging_helper import logger_wraps
from loguru import logger as LOGGER

import dt_tools.net.ip_info_helper as ih
from dt_tools.net import net_helper as nh
from dt_tools.console.console_helper import ConsoleHelper as console
from dt_tools.console.console_helper import TextStyle


BASE_URL='https://ipinfo.io'
TOKEN="NOT SET"

IP_INFO_TOKEN_LOCATION=pathlib.Path('~').expanduser().absolute() / ".IpHelper" / "ipinfo_token.json"
IP_INFO_CACHE_LOCATION=pathlib.Path('~').expanduser().absolute() / ".IpHelper" / "cache.json"
MAC_INFO_LOCATION=pathlib.Path('~').expanduser().absolute() / ".IpHelper" / "mac_info.json"

_SECONDS_IN_MINUTE = 60
_SECONDS_IN_HOUR = 60 * _SECONDS_IN_MINUTE
_CACHE_TTL_HOURS = 48 
_CACHE_SEMAPHORE = Semaphore()
_UNKNOWN = 'unknown'

class IpHelper():
    """
    This class provides information about an IP address.  
    
    It interfaces with the free **ipinfo.io** site.  The ipinfo.io site
    requires a user token which is free.

    - See 'setting up user token' in docs for information on aquiring and setting up token.

    In order to minimize API calls and improve performance, a cache
    for IP and MAC information is created and stored locally.
      
    The local data will be refresed if it is > 48 hours old. It can also
    be manually refreshed/cleared from cache.
    
    The class provides the following information:

    For WAN IPs:

    - ip         : xxx.xxx.xxx.xxx
    - hostname   : hostname.domain
    - city       : Atlanta
    - region     : Georgia
    - country    : US
    - loc        : 33.7490,-84.3880
    - org        : XXXXXXXXXXXX
    - postal     : 30302
    - timezone   : America/New_York

    For LAN IPs:

    - ip         : xxx.xxx.xxx.xxx
    - hostname   : hostname.domain[or workgroup]
    - bogon      : True (identifies this as a local IP)
    - mac        : XX:XX:XX:XX:XX:XX
    - vendor     : Raspberry Pi Trading Ltd

    """
    # Class variables
    _cache_last_update_mtime: float = 0.0
    _cache: Dict[str, dict] = None
    _mac_info: Dict[str, dict] = None  # Manually maintained.
    _token_initialized: bool = False

    @classmethod
    @logger_wraps(level="TRACE")
    def __init__(cls, purge_stale_entries: bool = True, no_token:bool = False):
        if not no_token:
            cls._validate_token()        
        cls._load_cache(purge_stale_entries)

    @property
    def cache_dict(cls) -> dict:
        """
        The IP Info cache dictionary.
        """
        return IpHelper._cache
            
    @classmethod            
    @logger_wraps(level="TRACE")
    def clear_cache(cls, ip_address: str = None)-> int:
        """
        Clear the IP Info cache or cache entry.

        Keyword Arguments:
            ip_address -- IP address to be cleared (default: {None})
              If None, all entries will be removed.
        
        Returns:
            int -- Number of entries removed.
        """
        IP_INFO_CACHE_LOCATION.parent.mkdir(exist_ok=True)
        IP_INFO_CACHE_LOCATION.touch(exist_ok=True)
        cls._load_cache()
        initial_cache_len = len(IpHelper._cache)
        if ip_address:
            if ip_address in IpHelper._cache.keys():
                del IpHelper._cache[ip_address]
                _CACHE_SEMAPHORE.acquire(timeout=2.0)
                IP_INFO_CACHE_LOCATION.write_text(json.dumps(IpHelper._cache))
                _CACHE_SEMAPHORE.release()
                IpHelper._cache_last_update_mtime = IP_INFO_CACHE_LOCATION.stat().st_mtime
                LOGGER.debug(f'{ip_address} removed from cache')
            else:
                LOGGER.warning(f'{ip_address} does NOT exist in cache')
        else:
            _CACHE_SEMAPHORE.acquire(timeout=2.0)
            IP_INFO_CACHE_LOCATION.write_text("{}")
            _CACHE_SEMAPHORE.release()
            IpHelper._cache = {}
            IpHelper._cache_last_update_mtime = IP_INFO_CACHE_LOCATION.stat().st_mtime
        
        cur_len = len(IpHelper._cache)
        removed_cnt = initial_cache_len - cur_len
        LOGGER.debug(f'IpHelper Cache cleared.  {removed_cnt} entries removed. {cur_len} entries remaining.')
        return removed_cnt
    
    @classmethod
    def is_cached(cls, ip_address: str) -> bool:
        """
        Check if IP address is in cache.

        Arguments:
            ip_address -- Ip address to be checked.

        Returns:
            True if found else False.
        """
        cls._load_cache()  # Load cache if needed
        ip_info = IpHelper._cache.get(ip_address, None)
        if ip_info:
            return ip_info.get('_cached', False)
    
        return False
    
    @classmethod
    @logger_wraps(level="TRACE")
    def get_ip_info(cls, ip_address: str, 
                    include_unknown_fields: bool = False, 
                    include_private_fields: bool = False,
                    bypass_cache: bool = False) -> dict:
        """
        Get all information related to IP Address

        Arguments:
            ip_address: Target IP address.

        Keyword Arguments:
            bypass_cache: If true, force API call and refresh (default: {False}).

        Returns:
            JSON dictionary of all data related to target IP address.
        """
        urllib3.disable_warnings()
        cls._load_cache() 
        entry_updated = False
        if not nh.is_ipv4_address(ip_address):
            ip_info = {'ip': ip_address, "title": "Not IPv4 address", "error": "Only IPv4 addresses supported"}
            return ip_info

        LOGGER.debug(f'GET_IP_INFO for {ip_address}')   
        if bypass_cache:
            LOGGER.debug('- Bypassing cache lookup')
            ip_info = None
        else:
            ip_info = IpHelper._cache.get(ip_address, None)
            if ip_info and cls._stale(ip_address):
                LOGGER.debug(f'- {ip_address} stale, will re-fresh')
                ip_info = None

        if ip_info is None:
            LOGGER.debug(f'- {ip_address} NOT in cache or stale')
            url=f'{BASE_URL}/{ip_address}?token={TOKEN}'
            try:
                resp = requests.get(url, verify=False)
                ip_info = resp.json()
                entry_updated = True
            except Exception as ex:
                resp = ""
                LOGGER.debug(f'  ERROR- url: {url} ex: {repr(ex)}')
                ip_info = {'ip': {ip_address}, "title": "Error in API call", 'error': f'Exception: {url}- {repr(ex)}'}

        if 'error' in ip_info.keys():
            LOGGER.warning(f'ERROR - url: {url}  resp: {ip_info}')
            return ip_info

        ip = ip_info['ip']
        mac = ip_info.get('mac', _UNKNOWN)
        if ip_info.get('hostname', None) is None:
            LOGGER.debug(f'- get hostname from ip {ip}')
            ip_info['hostname'] = nh.get_hostname_from_ip(ip)
            if _UNKNOWN not in ip_info['hostname']:        
                entry_updated = True

        if _UNKNOWN in mac:
            LOGGER.debug(f'- get mac address for ip {ip}')
            mac = nh.get_mac_address(ip)
            if mac is None:
                mac = _UNKNOWN
            else:
                ip_info['vendor'] = nh.get_vendor_from_mac(mac)
                entry_updated = True

        if _UNKNOWN in mac:
            LOGGER.debug(f'- Unable to determine MAC from ip {ip}')
        else:
            ip_info['mac'] = mac
            if _UNKNOWN in ip_info['hostname']:
                # Check the Mac Info cache
                hostname = IpHelper._mac_info.get(mac,{'hostname': _UNKNOWN})['hostname']
                LOGGER.debug(f'- {mac} - Lookup hostname based on mac_info cache. [{hostname}]')
                ip_info['hostname'] = f'-> {hostname}'
                if _UNKNOWN not in hostname:
                    entry_updated = True
            if _UNKNOWN in ip_info.get('vendor',_UNKNOWN):
                # Check the Mac Info cache
                vendor = IpHelper._mac_info.get(mac,{'vendor': _UNKNOWN})['vendor']
                LOGGER.debug(f'- {mac} - Lookup vendor based on mac_info cache. {vendor}')
                ip_info['vendor'] = f'-> {vendor}'
                if _UNKNOWN not in vendor:
                    entry_updated = True

        if entry_updated:
            ip_info["_cached"] = datetime.now().isoformat() 
            cls._load_cache() 
            _CACHE_SEMAPHORE.acquire(timeout=5.0)
            IpHelper._cache[ip_address] = ip_info
            # TODO: Performance, ONLY write new entry
            IP_INFO_CACHE_LOCATION.write_text(json.dumps(IpHelper._cache))
            _CACHE_SEMAPHORE.release()
            LOGGER.debug(f'SUCCESS: cache updated: {ip}/{ip_info["hostname"]}/{mac}')
            IpHelper._cache_last_update_mtime = IP_INFO_CACHE_LOCATION.stat().st_mtime
        
        if not include_private_fields or not include_unknown_fields:
            new_dict = {}
            for key, value in ip_info.items():
                if not include_private_fields and key.startswith('_'):
                    continue
                if not include_unknown_fields and value == _UNKNOWN:
                    continue
                new_dict[key] = value
            ip_info = new_dict

        return ip_info

    @classmethod
    def get_wan_ip_info(cls) -> Tuple[dict, int]:
        """
        Return the WAN (ie. Internet) address for this device.

        Returns:
            Tuple[dict, int]: ip_info entry and return code.
        """
        urllib3.disable_warnings()
        cls._validate_token()
        ip_info = None
        url=f'{BASE_URL}?token={TOKEN}'
        resp = ""
        rc = -1
        try:
            resp = requests.get(url, verify=False)
            ip_info = resp.json()
            rc = resp.status_code
        except Exception as ex:
            ip_info = {'error': f'Exception: {repr(ex)}'}

        if 'error' in ip_info:
            LOGGER.warning(f'ERROR- url: {url}  resp: {ip_info}')

        return (ip_info, rc)
        
    @classmethod        
    def find_in_cache(cls, search_token: str) -> List[str]:
        """
        Search each cache entry for supplied search string.

        Arguments:
            search_token: Text to search for.

        Returns:
            A list of IPs (str) whose data containes the search_token.
        """
        cls._load_cache() # update memory with any changes
        found_keys = []
        for key, entry in IpHelper._cache.items():
            for field_val in entry.values():
                if search_token.lower() in str(field_val).lower():
                    found_keys.append(key)
                    LOGGER.debug(f'- found {key} in {field_val}')
                    break

        found = len(found_keys)
        if found == 0:
            LOGGER.warning(f'- Search key [{search_token}] NOT found')
        else:
            for key in found_keys:
                cls.list_cache(key)
            LOGGER.info('')
            LOGGER.success(f'- {found} entries found.')

        return found_keys
        
    @classmethod
    def list_cache(cls, ip: str = None, show_all_fields: bool = True):
        """
        Print formatted contents of cache to console.

        Args:
            ip (str, optional): Target IP address. Defaults to None.
            show_all_fields (bool, optional): Include private (_) fields. Defaults to True.
        """        
        cls._load_cache() # Update memory with any changes
        if len(IpHelper._cache) == 0:
            LOGGER.warning('Sorry, cache is empty')
        elif ip is not None:
            entry = IpHelper._cache.get(ip, None)
            if entry:
                LOGGER.success(f'{console.cwrap(f"-- {ip} ---------------------------------------", style=TextStyle.BOLD)}')
                cls._print_entry(entry, show_all=show_all_fields)
            else:
                LOGGER.warning(f'{ip} does NOT exist in cache.')
        else:
            for key, entry in IpHelper._cache.items():
                LOGGER.success(f'{console.cwrap(f"-- {key} ---------------------------------------", style=TextStyle.BOLD)}')
                cls._print_entry(entry, show_all=show_all_fields)
                LOGGER.info('')
            LOGGER.info(f'IP Info cache contains {len(IpHelper._cache)} entries')

    @classmethod
    def _validate_token(cls) -> bool:
        LOGGER.debug('Token validation in progress...')
        rc = -1
        if not cls._token_initialized:
            if not IP_INFO_TOKEN_LOCATION.exists():
                LOGGER.error('Cached token NOT found.')
                cls._raise_token_error()            
            token_dict: dict = json.loads(IP_INFO_TOKEN_LOCATION.read_text())
            ih.TOKEN = token_dict.get('token', 'NOT VALID')
            cls._token_initialized = True
            ip_info, rc = cls.get_wan_ip_info()
            if rc != 200:
                LOGGER.error(f'Token check returns: {(ip_info)}')
                cls._raise_token_error()

        return True

    @classmethod
    def _raise_token_error(cls):
        LOGGER.warning('A token is required for IpHelper to function.')
        LOGGER.warning('Tokens are free, to get your token, go to "https://ipinfo.io/missingauth')
        LOGGER.warning('When you have a token run dt_set_iptoken to set the token for IpHelper.')
        raise RuntimeError('Invalid token for ipinfo.io.  See log and https://ipinfo.io/missingauth')

    @classmethod        
    def _load_cache(cls, purge_stale: bool = False):
        _CACHE_SEMAPHORE.acquire(timeout=15.0)
        if cls._refresh_required:
            IpHelper._cache = {}
            if not IP_INFO_CACHE_LOCATION.exists():
                cls.clear_cache()
            try:
                buffer = IP_INFO_CACHE_LOCATION.read_text()
                if buffer == "":
                    buffer = "{}"
                IpHelper._cache = json.loads(buffer)
                LOGGER.debug(f'{len(IpHelper._cache)} IpHelper cache entries loaded.')
                if purge_stale:
                    dropped = cls._drop_stale_entries()
                    LOGGER.debug(f'{dropped} IpHelper cache stale entries dropped.')
                IpHelper._cache_last_update_mtime = IP_INFO_CACHE_LOCATION.stat().st_mtime
            except Exception as ex:
                LOGGER.error(f'Unable to load IP Info Cache: {IP_INFO_CACHE_LOCATION}')
                LOGGER.warning(f'  {repr(ex)}')

        if IpHelper._mac_info is None:
            if not MAC_INFO_LOCATION.parent.exists():
                LOGGER.debug(f'Create directory {MAC_INFO_LOCATION.parent}')
                MAC_INFO_LOCATION.parent.mkdir(exist_ok=True)
            if not MAC_INFO_LOCATION.exists():
                LOGGER.debug(f'Create MAC Cache file: {MAC_INFO_LOCATION}')
                MAC_INFO_LOCATION.touch(exist_ok=True)
            buffer = MAC_INFO_LOCATION.read_text()
            if buffer == "":
                buffer = "{}"
            LOGGER.debug(f'Load MAC info from {MAC_INFO_LOCATION}')
            IpHelper._mac_info = json.loads(buffer)

        _CACHE_SEMAPHORE.release()

    @property
    def _refresh_required(cls) -> bool:
        refresh = False
        if IP_INFO_CACHE_LOCATION.exists():
            last_update = IP_INFO_CACHE_LOCATION.stat().st_mtime
            if last_update > IpHelper._cache_last_update_mtime:
                # Cache has been updated
                LOGGER.debug('Cache refresh required!')
                refresh = True
        elif IpHelper._cache is None:
            LOGGER.debug('Cache is empty, refresh required!')
            refresh = True
        else:
            LOGGER.debug('Cache refresh is NOT required.')
        return refresh

    @classmethod
    def _stale(self, ip_address: str) -> bool:
        """Return true if cached entry is stale"""
        ip_info = IpHelper._cache.get(ip_address, None)
        if ip_info:
            cached_time = ip_info.get('_cached', None)
            if cached_time:
                created = parser.parse(cached_time)
                delta = datetime.now() - created
                age_in_hours = delta.total_seconds() / _SECONDS_IN_HOUR
                # LOGGER.debug(f'{ip_info["hostname"]} age in hours {age_in_hours}')
                if age_in_hours > _CACHE_TTL_HOURS:
                    return True
            
        return False

    @classmethod
    def _age(cls, entry: dict) -> str:
        cached_time =  entry.get('_cached', None)
        age = 'N/A'
        if cached_time:
            created = parser.parse(cached_time)
            td = datetime.now() - created
            age = f'{td.days} days, {td.seconds//3600} hours, {(td.seconds//60)%60} minutes'
        return age

    @classmethod
    def _expires(cls, entry: dict) -> str:
        cached_time =  entry.get('_cached', None)
        expires = "Unknown"
        ttl = 'N/A'
        if cached_time:
            created = parser.parse(cached_time)
            expire_time = created + timedelta(hours=_CACHE_TTL_HOURS)
            ttl = expire_time - datetime.now()
            if ttl.seconds < 0:
                expires = "expired."
            else:
                expires = f'{ttl.days} days, {ttl.seconds//3600} hours, {(ttl.seconds//60)%60} minutes'
        return expires

    @classmethod
    def _drop_stale_entries(self) -> int:
        dropped = 0
        del_list = []
        for ip_address in IpHelper._cache.keys():
            # Create list of IPs to be removed
            if self._stale(ip_address):
                del_list.append(ip_address)
        for ip_address in del_list:
            # Remove identified IPs
            self.clear_cache(ip_address)
            dropped += 1
        return dropped 
    
    @classmethod
    def _print_entry(cls, entry: dict, show_all: bool = True):
        entry['_cache_age'] = cls._age(entry)
        entry['_expires'] = cls._expires(entry)
        entry['_stale'] = cls._stale(entry['ip'])
        for k,v in entry.items():
            if not show_all and k.startswith('_'):
                pass
            else:
                LOGGER.info(f'  {k:10} : {v}')




"""
This module creates the token file and stores the token used for interface with ipinfo.io.

To get your token, go to https:/ipinfo.io/missingauth

```
poetry run python -m dt_tools.cli.set_iphelper_token_cli
```

"""
import json

import dt_tools.logger.logging_helper as lh
from dt_tools.console.console_helper import (ColorFG, TextStyle,
                                             ConsoleHelper, ConsoleInputHelper)
from loguru import logger as LOGGER 

import dt_tools.net.ip_info_helper as ih


def manage_token():
    console = ConsoleHelper()
    console_input = ConsoleInputHelper()

    ip_helper = console.cwrap("IpHelper", fg=ColorFG.WHITE2, style=TextStyle.BOLD)
    note = console.cwrap('NOTE:', fg=ColorFG.YELLOW2, style=[TextStyle.BOLD, TextStyle.ITALIC])
    token_file = console.cwrap(ih.IP_INFO_TOKEN_LOCATION, fg=ColorFG.WHITE2, style=TextStyle.BOLD)

    console.print('')
    console.print_line_separator('', 90)
    console.print_line_separator(' IpHelper Token Manager', 90)
    console.print('')
    console.print(f'The {ip_helper} tools and packages (dt-cli-tools, dt-net) need a valid token from ipinfo.io.')
    console.print('  (see https://ipinfo.io/missingauth)')
    console.print('')
    console.print('To enable the tools and packages, a one-time process is necessary to acquire the API token, then')
    console.print('save it locally for future use.')
    console.print('')
    console.print('Once you have aquired a valid token, it may be entered via this script. The token will')
    console.print('saved and made available for the IpHelper routines.')
    console.print('')
    console.print('If you already have a token, but forget what it is, you may log back into ipinfo.io')
    console.print('and retrieve your token.')
    console.print('')
    console.print(f'{note}  The token is stored locally in {token_file}. x')
    console.print( '         format: {"token": "xxxxxxxxxxxxxx"}')
    console.print('')

    if console_input.get_input_with_timeout('Continue (y/n) > ', ['y', 'n']) == 'y':
        token = console_input.get_input_with_timeout('Token > ')
        if len(token.strip()) == 0:
            LOGGER.warning('Empty token, did not save.')
        else:
            token_dict = json.dumps({"token": token})
            ih.IP_INFO_TOKEN_LOCATION.parent.mkdir(parents=True, exist_ok=True)
            ih.IP_INFO_TOKEN_LOCATION.write_text(token_dict)
            LOGGER.success('Token saved.')
        LOGGER.info('')
    
    print('Process complete.')

if __name__ == "__main__":
    lh.configure_logger()
    manage_token()



from typing import List

from loguru import logger as LOGGER

import dt_tools.console.console_helper as ch
import dt_tools.logger.logging_helper as lh
import dt_tools.net.net_helper as helper
from dt_tools.console.console_helper import ColorFG, ConsoleHelper
from dt_tools.console.console_helper import ConsoleInputHelper as con_input
from dt_tools.console.spinner import Spinner, SpinnerType
from dt_tools.net.net_helper import LAN_Client

# COMMON_PORTS = {
#     "Echo service": 7,
#     "FTP-data": 20,
#     "FTP": 21,
#     "SSH": 22,
#     "Telnet": 23,
#     "SMTP": 25,
#     "DNS": 53,
#     "TFTP": 69,
#     "HTTP": 80,
#     "Kerberos": 88,
#     "Iso-tsap": 102,
#     "POP3": 110,
#     "MS EPMAP": 135,
#     "NetBIOS-ns": 137,
#     "NetBIOS-ssn": 139,
#     "IMAP4": 143,
#     "HP Openview (alarm)": 381,
#     "HP Openview (data)": 383,
#     "HTTPS": 443,
#     "Kerberos (pwd)": 464,
#     "SMTP TLS/SSL": 465,
#     "SMTP (submission)": 587,
#     "MS DCOM": 593,
#     "LDAP TLS/SSL": 636,
#     "MS Exchange": 691,
#     "VMWare": 902,
#     "FTP SSL (data)": 989,
#     "FTP SSL (control)": 990,
#     "IMAP4 SSL": 993,
#     "POP3 SSL": 995,
#     "MS RPC": 1025,
#     "OpenVPN": 1194,
#     "WASTE": 1337,
#     "Cisco VQP": 1589,
#     "Steam": 1725,
#     "cPanel": 2082,
#     "radsec": 2083,
#     "Oracle DB": 2483,
#     "Oracle DB SSL": 2484,
#     "Semantec AV": 2967,
#     "XBOX Live": 3074,
#     "MySQL": 3306,
#     "World of Warcraft": 3724,
#     "Google Desktop": 4664,
#     "PostgresSQL": 5432,
#     "RFB/VNC": 5900,
#     "IRC1": 6665,
#     "IRC2": 6666,
#     "IRC3": 6667,
#     "IRC4": 6668,
#     "IRC5": 6669,
#     "BitTorrent": 6881,
#     "Quicktime": 6970,
#     "BitTorrent2": 6999,
#     "Kaspersky CC": 8086,
#     "Kaspersky": 8087,
#     "VMWare Server": 8222,
#     "PDL": 9100,
#     "BackupExec": 10000,
#     "NetBus": 12345,
#     "Sub7": 27374,
#     "Back Orifice": 31337,
# }

def display_LAN_report():
    ch = ConsoleHelper()
    ch.print_line_separator('LAN Client Report', 40)
    
    spinner = Spinner("Retrieve ARP Entries", SpinnerType.ARC, show_elapsed=True)

    spinner.start_spinner('via ARP Cache')
    c_clients = helper.get_lan_clients_from_ARP_cache(include_hostname=True, include_mac_vendor=True)
    spinner.stop_spinner()
    LOGGER.info(f'ARP Cache     returns {len(c_clients)} devices.')

    spinner.start_spinner('via ARP Broadcast')
    b_clients = helper.get_lan_clients_ARP_broadcast(include_hostname=True, include_mac_vendor=True)
    spinner.stop_spinner()
    LOGGER.info(f'ARP Broadcast returns {len(b_clients)} devices.')
    LOGGER.info('')
    
    b_idx = 0
    b_done = False
    c_idx = 0
    c_done = False
    # Sort lists so we can match results
    sb_clients = sorted(b_clients, key=lambda x: x.mac)
    sc_clients = sorted(c_clients, key=lambda x: x.mac)
    
    print(f'{"ARP Broadcast":17}  {"ARP Cache":17}  {"ip":15} {"Hostname":33} {"Vendor":25}')
    print(f'{"-"*17}  {"-"*17}  {"-"*15} {"-"*33} {"-"*25}')
    while not b_done and not c_done:
        broadcast_entry = sb_clients[b_idx]
        cache_entry = sc_clients[c_idx]
        if broadcast_entry.mac == cache_entry.mac:
            b_mac = broadcast_entry.mac
            c_mac = cache_entry.mac
            ip = broadcast_entry.ip
            hostname = broadcast_entry.hostname
            vendor = broadcast_entry.vendor
            b_idx += 1
            c_idx += 1
        elif broadcast_entry.mac < cache_entry.mac:
            b_mac = broadcast_entry.mac
            c_mac = ""
            ip = broadcast_entry.ip
            hostname = broadcast_entry.hostname
            vendor = broadcast_entry.vendor
            b_idx += 1
        else:
            b_mac = ""
            c_mac = cache_entry.mac
            ip = cache_entry.ip
            hostname = cache_entry.hostname if cache_entry.hostname is not None else '?'
            vendor = cache_entry.vendor if cache_entry.vendor is not None else '?'
            c_idx += 1
        try:
            print(f'{b_mac:17}  {c_mac:17}  {ip:15} {hostname:33} {vendor}')
        except TypeError:
            print(f'{b_mac    =}')
            print(f'{c_mac    =}')
            print(f'{ip       =}')
            print(f'{hostname =}')
            print(f'{vendor   =}')
        if b_idx == len(sb_clients):
            b_done = True
        if c_idx == len(sc_clients):
            c_done = True
        b_idx = min(b_idx, len(sb_clients)-1)
        c_idx = min(c_idx, len(sc_clients)-1)

def demo():
    ConsoleHelper.print('')
    ConsoleHelper.print_line_separator('', 80)
    ConsoleHelper.print_line_separator('dt_net_helper_demo', 80)
    ConsoleHelper.print('')

    # Get local machines internal IP
    local_ip = helper.get_local_ip()
    # Get local machines External IP
    wan_ip = helper.get_wan_ip()
    # Get list of client machines on LAN
    lan_list: List[LAN_Client] = helper.get_lan_clients_from_ARP_cache()
    ip_dict = {"Local IP": local_ip, "WAN IP": wan_ip, "Bad IP": "192.168.1.0" }
    if len(lan_list) > 3:
        # Choose two clients
        ip_dict['Client1'] = lan_list[0].ip
        ip_dict['Client2'] = lan_list[len(lan_list)-1].ip

    # Display information for each client (machine) in list
    for ip_name, ip in ip_dict.items():
        is_valid = ConsoleHelper.cwrap('Valid', ColorFG.GREEN2) if helper.is_valid_host(ip) else ConsoleHelper.cwrap('Invalid', ColorFG.RED2)
        ip_type = 'Unknown'
        if helper.is_ipv4_address(ip):
            ip_type = "IPv4"
        elif helper.is_ipv6_address(ip):
            ip_type = "IPv6"

        if is_valid == 'Valid':
            hostname = helper.get_hostname_from_ip(ip)
            mac = helper.get_mac_address(ip)
            vendor = helper.get_vendor_from_mac(mac) if mac is not None else 'unknown'
            is_alive = helper.ping(ip)

        ConsoleHelper.print_line_separator(f'{ip_name} Info', 40)
        ConsoleHelper.print(f'IP         : {ConsoleHelper.cwrap(ip,ColorFG.YELLOW)} {is_valid}')
        ConsoleHelper.print(f'IP Type    : {ConsoleHelper.cwrap(ip_type,ColorFG.YELLOW)}')
        if is_valid == 'Valid':
            ConsoleHelper.print(f'hostname   : {ConsoleHelper.cwrap(hostname,ColorFG.YELLOW)}')
            ConsoleHelper.print(f'mac        : {ConsoleHelper.cwrap(mac,ColorFG.YELLOW)}')
            ConsoleHelper.print(f'mac vendor : {ConsoleHelper.cwrap(vendor,ColorFG.YELLOW)}')
            ConsoleHelper.print(f'is alive   : {ConsoleHelper.cwrap(is_alive,ColorFG.YELLOW)}')
            if is_alive:
                resp = con_input.get_input_with_timeout('Perform a port scan (y/n)? ', ['y', 'n'], default='n', timeout_secs=5)
                ConsoleHelper.cursor_up()
                ConsoleHelper.clear_line()
                if resp == 'y':
                    ConsoleHelper.print_line_separator('Scan for open Ports: ', 27)
                    ConsoleHelper.print('  .', eol='')
                    for port_use, port in helper.COMMON_PORTS.items():
                        if helper.is_port_open(ip, port):
                            ConsoleHelper.cursor_move(column=1)
                            ConsoleHelper.clear_to_EOL()
                            ConsoleHelper.print(f'  {port_use:12} : {port} open')
                            ConsoleHelper.print('  .', eol='')
                        else:
                            ConsoleHelper.print('.',eol='')
                    ConsoleHelper.cursor_move(column=1)
                    ConsoleHelper.clear_to_EOL()
        ConsoleHelper.print('')

    display_LAN_report()

if __name__ == '__main__':
    ch.enable_ctrl_c_handler()
    lh.configure_logger()
    demo()
    


"""
Wake-on-LAN utility class

This class can be used to send WOL packet to target machines via IP or Hostname.

The wol function can optionally wait for the host to 'wake-up' and provide a
status message indicating success or failure.

Raises:
    ex: Invalid MAC address
    ValueError: When MAC address does not conform to expected format

Example::

    from dt_tools.net.wol import WOL

    wol = WOL()
    wol.send_wol_to_host(myTargetHost, wait_secs=60)
    print(wol.status_message)



"""
import socket
import struct
import time

from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
import dt_tools.net.net_helper as nh

# from dt_tools.console.progress_bar import ProgressBar

_LOGLEVEL="INFO"

class WOL():
    '''
    Send WOL request to target device via MAC address, hostname or IP
    and optionally, wait for device to come online.

    Example::

        from dt_tools.net.wol import WOL

        wol = WOL()
        wol.send_wol_to_host('somehost', wait_secs=20)
        print(wol.status_message)
    '''
    _BROADCAST_IP = "255.255.255.255"
    _DEFAULT_PORT = 9
    
    _status_message: str = ''

    def __init__(cls):
        LOGGER.debug('WOL class created.')

    @property
    def status_message(cls) -> str:
        """Return WOL request status message"""
        return cls._status_message
    
    def send_wol_to_host(cls, host_name: str, wait_secs: int = 0) -> bool:
        """
        Send WOL to target hostname or IP

        Args:
            host_name (str): host name (or IP Address).
            wait_secs (int, optional): Number of secs to wait for host to come online. Defaults to 0.

        Returns:
            bool: True if wol packet successfully sent, else False
        """
        LOGGER.debug(f'send_wol_via_host("{host_name}")')
        return_status = False
        try:
            if nh.is_valid_ipaddress(host_name):
                _ip = host_name
                _host = nh.get_hostname_from_ip(_ip)
            else:
                _ip = nh.get_ip_from_hostname(host_name)
                _host = host_name
            if not nh.is_valid_host(host_name):
                host_type = 'ip address' if _host == nh._UNKNOWN else 'hostname'
                cls._status_message = f'Invalid {host_type} [{host_name}]'
            else:
                mac = nh.get_mac_address(_ip)
                if mac:
                    return cls.send_wol_via_mac(mac, wait_secs=wait_secs, ip=_ip)  
                cls._status_message = f"Unable to determine MAC address for {host_name}"
                LOGGER.debug(cls._status_message)
        except Exception as ex:
            LOGGER.exception(repr(ex))
            cls._status_message = f'Unable to send wol packet to {hostname} - {repr(ex)}.'

        return return_status
    
    # === Private functions ==============================================================
    def send_wol_via_mac(cls, mac: str, wait_secs: int = 0, ip: str = None) -> bool:
        LOGGER.debug(f'send_wol_via_mac("{mac}")')
        try:
            mac_address = nh.unformat_mac(mac)
        except ValueError as ve:
            cls._status_message = f'Incorrectly formatted MAC address: {mac}'
            LOGGER.error(f'{cls._status_message} - {repr(ve)}')
            return False
        # macAddress = mac
        # # Check macaddress format and try to compensate.
        # if len(macAddress) == 12:
        #     pass
        # elif len(macAddress) == 12 + 5:
        #     sep = macAddress[2]
        #     macAddress = macAddress.replace(sep, "")
        # else:
        #     cls._status_message = f"Incorrect MAC address format: {macAddress}"
        #     return False

        # Pad the synchronization stream.
        LOGGER.debug(f'pad the mac address: {mac_address}')
        data = b"FFFFFFFFFFFF" + (mac_address * 20).encode()
        send_data = b""
        LOGGER.debug(f'padded  mac address: {data}')
        # Split up the hex values and pack.
        for i in range(0, len(data), 2):
            send_data += struct.pack("B", int(data[i : i + 2], 16))

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        LOGGER.debug(f'Sending WOL to {mac}')
        try:
            sock.sendto(send_data, ("<broadcast>", 9))  # 7
        except Exception as ex:
            LOGGER.warning(f'Error sending WOL: {repr(ex)}')
        finally:
            sock.close()

        if wait_secs == 0:
            cls._status_message = "Packet sent."
            return True

        if ip is None:
            ip = nh.get_ip_from_mac(mac)
            if ip is None:
                LOGGER.debug(f'ERROR: Unable to resolve IP from MAC {mac}')
                cls._status_message = "Packet sent, no wait, could not resolve IP."
                return True

        is_online = cls._wait_for_device_to_come_online(ip, wait_secs)
        hostname = nh.get_hostname_from_ip(ip)
        if hostname:
            LOGGER.debug(f'Host: {hostname} | MAC: {mac} | {ip} - is online {is_online}')
        else:
            LOGGER.debug(f'MAC: {mac} | {ip} - is online {is_online}')
        
        cls._status_message = "Packet sent.  Host online." if is_online else f"Packet sent. Host NOT online after waiting {wait_secs} seconds."
        return is_online

    def _create_magic_packet(cls, mac: str) -> bytes:
        if len(mac) == 17:
            sep = mac[2]
            mac = mac.replace(sep, "")
        elif len(mac) != 12:
            raise ValueError("Incorrect MAC address format")

        return bytes.fromhex("F" * 12 + mac * 16)
                    
    def _wait_for_device_to_come_online(cls, ip: str, wait_secs: int) -> bool:
        is_online = nh.ping(ip)
        if not is_online:
            LOGGER.debug(f'Waiting for {wait_secs} seconds for device to come online.')
            start = time.time()
            elapsed = 0
            while elapsed < wait_secs and not is_online:
                time.sleep(1)
                is_online = nh.ping(ip)
                elapsed = time.time() - start

        return is_online


if __name__ == "__main__":
    lh.configure_logger(log_level="DEBUG", log_format=lh.DEFAULT_CONSOLE_LOGFMT)
    wol = WOL()
    hostnames = ['nirvana', 'badhost']
    for hostname in hostnames:
        LOGGER.info('')
        LOGGER.info(f'Sending WOL to {hostname}...')
        try:
            if wol.send_wol_to_host(hostname, wait_secs=20):
                LOGGER.success(f'  {wol.status_message}')
            else:
                LOGGER.warning(f'  {wol.status_message}')
        except Exception as ex:
            LOGGER.error(repr(ex))



"""
Network utilities helper module.

Functions to assist with network related information and tasks.

- ping
- local IP address
- get ip for given hostname
- get hostname for given ip
- get mac address for given hostname or ip
- get mac vendor
- get local client info on LAN

"""
import ipaddress
import platform
import random
import socket
import subprocess
# import uuid
from dataclasses import dataclass
from time import sleep
from typing import List, Union

import requests
import scapy.all as scapy
from loguru import logger as LOGGER

from dt_tools.os.os_helper import OSHelper
from dt_tools.logger.logging_helper import logger_wraps

_UNKNOWN = 'unknown'

COMMON_PORTS = {
    "Echo service": 7,
    "FTP-data": 20,
    "FTP": 21,
    "SSH": 22,
    "Telnet": 23,
    "SMTP": 25,
    "DNS": 53,
    "TFTP": 69,
    "HTTP": 80,
    "Kerberos": 88,
    "Iso-tsap": 102,
    "POP3": 110,
    "MS EPMAP": 135,
    "NetBIOS-ns": 137,
    "NetBIOS-ssn": 139,
    "IMAP4": 143,
    "HP Openview (alarm)": 381,
    "HP Openview (data)": 383,
    "HTTPS": 443,
    "Kerberos (pwd)": 464,
    "SMTP TLS/SSL": 465,
    "SMTP (submission)": 587,
    "MS DCOM": 593,
    "LDAP TLS/SSL": 636,
    "MS Exchange": 691,
    "VMWare": 902,
    "FTP SSL (data)": 989,
    "FTP SSL (control)": 990,
    "IMAP4 SSL": 993,
    "POP3 SSL": 995,
    "MS RPC": 1025,
    "OpenVPN": 1194,
    "WASTE": 1337,
    "Cisco VQP": 1589,
    "Steam": 1725,
    "cPanel": 2082,
    "radsec": 2083,
    "Oracle DB": 2483,
    "Oracle DB SSL": 2484,
    "Semantec AV": 2967,
    "XBOX Live": 3074,
    "MySQL": 3306,
    "World of Warcraft": 3724,
    "Google Desktop": 4664,
    "PostgresSQL": 5432,
    "RFB/VNC": 5900,
    "IRC1": 6665,
    "IRC2": 6666,
    "IRC3": 6667,
    "IRC4": 6668,
    "IRC5": 6669,
    "BitTorrent": 6881,
    "Quicktime": 6970,
    "BitTorrent2": 6999,
    "Kaspersky CC": 8086,
    "Kaspersky": 8087,
    "VMWare Server": 8222,
    "PDL": 9100,
    "BackupExec": 10000,
    "NetBus": 12345,
    "Sub7": 27374,
    "Back Orifice": 31337,
}

@dataclass
class LAN_Client():
    """
    Data class to hold Lan Client information.

    """
    ip: str #: Device IP address
    mac: str #: Device MAC address
    hostname: str = None # Device hostname
    vendor: str = None # NIC Vendor

    def to_dict(self):
        """
        Return LAN_Client as a dictionary.
        """
        LAN_Client_dict = {}
        for attr in self.__dict__:
            LAN_Client_dict[attr] = self.__dict__[attr]

        return LAN_Client_dict        

# ===============================================================================================
def get_port_name(port:int) -> str:
    """
    Retrieve (common) port name if defined else none.

    Args:
        port (int): port number

    Returns:
        str: Port name if found, else none
    """
    port_name = None
    if port in COMMON_PORTS.values():
        for key, val in COMMON_PORTS.items():
            if val == port:
                port_name = key
                break
    return port_name

def _get_ipaddress_obj(ip: str) -> Union[ipaddress.IPv4Address, ipaddress.IPv6Address]:
    """
    get ip_address object

    Args:
        ip (str): String representation of IPv4 or IPv6 address.

    Returns:
        Union[ipaddress.IPv4Address, ipaddress.IPv6Address]: ipaddress object
    """
    ip_obj = None
    try:
        ip_obj = ipaddress.ip_address(ip)
    except ValueError:
        pass

    return ip_obj

def is_valid_ipaddress(ip: str) -> bool:
    """
    Check if IP is valid address.

    Args:
        ip (str): IP address.

    Returns:
        bool: True if valid else False.
    """
    return _get_ipaddress_obj(ip) is not None

def is_ipv4_address(ip: str) -> bool:
    """
    Validate IPv4 address in dotted-quad notation.
    
    Args:
        ip_address: (str) in format 999.999.999.999

    Example::

        >>> is_ipv4_address("1.2.3.4")
        True
        >>> is_ipv4_address("127.0.0.1/8")
        False
        >>> is_ipv4_address("1.2.3.4.5")
        False
    """
    ip_obj = _get_ipaddress_obj(ip)
    return ip_obj is not None and isinstance(ip_obj, ipaddress.IPv4Address)

def is_ipv6_address(ip: str) -> bool:
    """
    Validate IPv6 address.

    Args:
        ip (str): IPv6 formatted address.

    Returns:
        bool: True if valid else False.
    """
    ip_obj = _get_ipaddress_obj(ip)
    return ip_obj is not None and isinstance(ip_obj, ipaddress.IPv6Address)


def is_port_open(host_name: str, port: int, timeout:float=1.0) -> bool:
    """
    Check if port is open on target host.

    Args:
        host_name (str): Target host name
        port (int): Port number to test
        timeout (float, optional): Seconds to wait for connection. Defaults to 1.0.

    Returns:
        bool: True if port is open else False
    """
    port_is_open = False
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(timeout)
    LOGGER.debug(f'is_port_open() - attempting to connect to {host_name}:{port}')
    try:
        s.connect((host_name, int(port)))
        s.shutdown(socket.SHUT_WR)
        LOGGER.debug('is_port_open() - connection successful.')
        port_is_open = True
    except:  # noqa: E722
        LOGGER.debug(f'is_port_open() - unable to connect to {host_name}:{port}')
    finally:
        s.close()

    return port_is_open

@logger_wraps(level="TRACE")
def is_valid_host(host_name: str) -> bool:
    """
    Check if host_name is valid.  
    
    Args:
        host (str): Host name or IP address.

    Returns:
        bool: True if hostname is resolvable else False
    """
    # valid_host = False
    # try:
    #     ip_addr = ipaddress.ip_address(host_name)
    # except ValueError as ve:
    #     LOGGER.debug(f'host_name: {host_name} is NOT a valid IP address. [{ve}]')

    try:
        if is_valid_ipaddress(host_name):
            LOGGER.debug(f'is_valid_host() - check via IP addr: {host_name}')
            _ = socket.gethostbyaddr(host_name)
        else:
            LOGGER.debug(f'is_valid_host() - check via hostname: {host_name}')
            _ = socket.gethostbyname(host_name)
        valid_host = True

    except (socket.gaierror, socket.herror):
        valid_host = False

    return valid_host
    

# == get hostname, ip, MAC, vendor routins ======================================================
@logger_wraps(level="TRACE")
def get_hostname_from_ip(ip: str) -> str:
    """
    Get hostname from IP address.

    Arguments:
        ip: IP address to get hostname for.

    Returns:
        Hostname if found else 'unknown' if not found or error.
    """
    hostname = _UNKNOWN
    if ip == get_local_ip():
        hostname = get_local_hostname()
    else:
        try:
            host = socket.gethostbyaddr(ip)
            hostname = host[0]
        except Exception as ex:
            LOGGER.debug(f'Unable to get_hostname_from_ip("{ip}") - {repr(ex)}')

    return hostname

@logger_wraps(level="TRACE")
def get_ip_from_hostname(host_name: str = None) -> str:
    """
    Get IP address from hostname.

    Arguments:
        host_name: Name of target host, if missing, use local hostname

    Returns:
        IP address if found or '' if not found or error.

    """
    if host_name is None:
        host_name = socket.gethostname()

    try:
        ip = socket.gethostbyname(host_name)
    except socket.gaierror:
        ip = ''
    return ip

@logger_wraps(level="TRACE")
def get_ip_from_mac(mac: str) -> str:
    """
    Get IP address based on MAC (via ARP)

    Args:
        mac (str): MAC address

    Raises:
        ValueError: Unknown MAC address format
        ValueError: Unable to determine IP from MAC address

    Returns:
        str: IP address
    """
    mac = format_mac(mac)

    cmd_list = _arp_entries_command().split()
    try:
        process_rslt = subprocess.run(cmd_list, capture_output=True)
        rslt = process_rslt.stdout.decode('utf-8').splitlines()
    except Exception as ex:
        LOGGER.error(f'get_ip_from_mac("{mac}"): unable to run {cmd_list}.  {repr(ex)}')
        rslt = ""

    LOGGER.debug(f'MAC: {mac}\nRESULT: {rslt}')
    arp = [token for token in rslt if mac in token]
    arp_line = '' if len(arp) == 0 else arp[0]
    LOGGER.debug(f'  arp line: {arp}')
    ip = None
    if len(arp_line) > 0:
        try:
            if platform.system() == "Windows":
                ip = " ".join(arp_line.split()).split()[0]
            else:
                # This used to be [2] for some reason...
                ip = " ".join(arp_line.split()).split()[0]
        except Exception as ex:
            LOGGER.error(f'  mac [{mac}]-problem resolving ip: {arp_line} - {repr(ex)}')

    if ip is not None:
        LOGGER.debug(f'  MAC {mac} resolves to {ip}')
        return ip
    
    raise ValueError(f'Unable to resolve IP for mac {mac}')

def get_wan_ip() -> str:
    """
    Get the WAN (ie. Internet) IP address for this device.

    Returns:
        str: WAN IP or 'Unknown'
    """
    from dt_tools.net.ip_info_helper import IpHelper

    ip_info, _ = IpHelper.get_wan_ip_info()
    return ip_info.get('ip', _UNKNOWN)

@logger_wraps(level="TRACE")
def get_local_ip() -> str:
    """
    Get local IP address

    Returns:
        Local machine IP address, or '127.0.0.1' if not found or error
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('192.255.255.255', 1))
        ip = s.getsockname()[0]
    except:  # noqa: E722
        ip = '127.0.0.1'
    finally:
        s.close()

    return ip

@logger_wraps(level="TRACE")
def get_local_hostname() -> str:
    """
    Get local hostname

    Returns:
        str: local host name or 'unknown'
    """
    hostname = _UNKNOWN
    try:
        hostname = socket.gethostname()
    except Exception as ex:
        LOGGER.debug(f'Unable to get_local_hostname() - {repr(ex)}')

    return hostname

@logger_wraps(level="TRACE")
def get_mac_address(ip: str) -> str:
    """
    Get mac address for specified IP

    Returns:
        str: MAC address or None if not found.
    """
    
    if ip == get_local_ip():
        import uuid
        mac = (':'.join(['{:02x}'.format((uuid.getnode() >> ele) & 0xff) for ele in range(0,8*6,8)][::-1]))    
    else:
        from scapy.layers.l2 import getmacbyip
        mac = getmacbyip(ip)
    
    if mac is not None:
        mac = str(mac).upper()
    return mac

@logger_wraps(level="TRACE")
def get_vendor_from_mac(mac: str) -> str:
    """
    Return the vendor name for specified MAC address

    Arguments:
        mac: Target MAC address

    Returns:
        Vendor name if found, else 'unknown'
    """
    if mac is None:
        raise ValueError('MAC address cannot be None.')
    vendor = _UNKNOWN
    # url = f'https://api.maclookup.app/v2/macs/{mac}'
    url = f'https://api.macvendors.com/{mac}'
    retry = 0
    RETRY_MAX = 15
    try:
        while retry < RETRY_MAX and vendor == _UNKNOWN:
            resp = requests.get(url)
            if resp.status_code == 200:
                # vendor = resp.json()['company']      # api.maclookup
                vendor = resp.text # api.macvendors.com
                if len(vendor) == 0:
                    LOGGER.debug(f'  ERROR: Vendor not found, {url}, {resp.text}')
                    vendor = "Not Found"
                if retry > 0:
                    LOGGER.debug(f'  SUCCESS. Retry succeeded, {vendor}, {url}')    
            elif resp.status_code == 429:  # You've been throttled
                retry += 1
                sleep_secs = random.uniform(0.5,3.5)
                LOGGER.debug(f'  WARNING: Throttle ({mac}) [{retry}]... {sleep_secs:1.2} {url}')
                sleep(sleep_secs) # Throttle (limit = 2 requests/second)
            else:
                LOGGER.debug(f'  ERROR: MAC Lookup resp: {resp.status_code}, {url}, {resp.text}')
                retry = RETRY_MAX

    except Exception as ex:
        LOGGER.debug(f'  ERROR: MAC Lookup error {url}: {repr(ex)}')
        vendor = _UNKNOWN
        
    if vendor == _UNKNOWN:
        LOGGER.debug(f'  ERROR: Unable to resolve, {mac}')

    return vendor


# == ARP Calls ===============================================================================================
@logger_wraps(level="TRACE")
def get_lan_clients_ARP_broadcast(include_hostname: bool = False, include_mac_vendor: bool = False) -> List[LAN_Client]:
    """
    Retrieve a list of LAN_Clients from the local network.

    The list of clients are retrieved via a Scapy ARP broadcast

    Keyword Arguments:
        include_hostname: Include hostname information (default: {False})
        include_mac_vendor: Include MAC vendor name (default: {False})

    Raises:
        PermissionError: Linux requires this call to be run as ROOT

    Returns:
        A list of :class:`~LAN_Client` entries 
    
    Note:
        including hostname and/or vendor will slow down process due to additional calls

    """
    if OSHelper.is_linux() and not OSHelper.is_linux_root():
        LOGGER.critical('You must be root on linux for ARP_Broadcast to work')
        raise PermissionError('Must be root')

    request = scapy.ARP()
    request.pdst = _get_target_protocol_address_pdst() # '192.168.1.1/24'
    broadcast = scapy.Ether() 
    
    broadcast.dst = 'ff:ff:ff:ff:ff:ff'    
    request_broadcast = broadcast / request 
    clients = scapy.srp(request_broadcast, timeout=3, verbose=False)[0] 

    lan_client_list: List[LAN_Client] = []
    for element in clients: 
        ip = element[1].psrc
        if _trackable_ip(ip):
            mac = element[1].hwsrc
            entry = LAN_Client(ip, mac.upper())
            lan_client_list.append(entry)
    
    if include_hostname or include_mac_vendor:
        lan_client_list = _get_hostname_and_or_vendor(lan_client_list, include_hostname, include_mac_vendor)

    return lan_client_list        

@logger_wraps(level="TRACE")
def get_lan_clients_from_ARP_cache(include_hostname: bool = False, include_mac_vendor: bool = False) -> List[LAN_Client]:
    """     
    Retrieve a list of LAN_Clients from the local network.

    The list of clients are retrieved via the ARP cache.

    Keyword Arguments:
        include_hostname: Include hostname information (default: {False})
        include_mac_vendor: Include MAC vendor name (default: {False})

    Returns:
        A list of :class:`~LAN_Client` entries 

    Note:
        including hostname and/or vendor will slow down process due to additional calls

    """
    arp_cmd = ["arp", "-a"] if OSHelper().is_windows() else ["arp", "-n"]
    LOGGER.debug(f'ARP command: {arp_cmd}')
    process_rslt = subprocess.run(arp_cmd, capture_output=True)
    result = process_rslt.stdout.decode('utf-8').splitlines()
    # Remove lines that are not IPs
    tokens =  [ line.strip() for line in result if line.count('.')==3 and not line.startswith('Interface')]

    lan_client_list = []
    for arp_entry in tokens:
        # If windows, response will be IP MAC TYPE, else IP TYPE MAC
        arp_field = arp_entry.split()
        ip = arp_field[0]
        if _trackable_ip(ip):
            mac = arp_field[1] if OSHelper().is_windows() else arp_field[2]
            mac = mac.replace('-',':').upper()
            entry = LAN_Client(ip, mac)
            lan_client_list.append(entry)

    if include_hostname or include_mac_vendor:
        lan_client_list = _get_hostname_and_or_vendor(lan_client_list, include_hostname, include_mac_vendor, bypass_cache=False)

    return lan_client_list

# ===========================================================================================================
def ping(host_name: str, wait_secs: int = 1) -> bool:
    """
    Ping target host (ip or name).

    Arguments:
        host_name: Target hostname (or ip address)

    Keyword Arguments:
        wait_secs: Number of seconds to wait for a reply (default: {1}).

    Returns:
        True if host responds, False if timeout or error.
    """
    # Token can be an IP or a hostname
    if OSHelper().is_windows():
        ping_cmd = ['ping', '-n', '1', '-w', str(wait_secs * 1000), host_name]
    else:
        ping_cmd = ['ping', '-c', '1', '-W', str(wait_secs), host_name]

    online_state = False
    ping_result = subprocess.run(ping_cmd, capture_output=True)
    if ping_result.returncode == 0 and 'unreachable' not in ping_result.stdout.decode('utf-8'):
        online_state = True
            
    return online_state

# == Private Methods ==========================================================================
def _trackable_ip(ip: str) -> bool:
    trackable = True
    if ip.startswith('224.') or \
        ip.startswith('239.255.250.250') or \
        ip.startswith('239.255.255.250') or \
        ip.endswith('.255'):
        trackable = False
    
    return trackable

@logger_wraps(level="TRACE")
def _get_hostname_and_or_vendor(client_list: list, include_hostname: bool, include_mac_vendor: bool, bypass_cache: bool = False) -> List[LAN_Client]:

    updated_list: List[LAN_Client] = []
    from dt_tools.net.ip_info_helper import IpHelper
    ip_info = IpHelper()
    arp_entry: LAN_Client = None

    for arp_entry in client_list:
        ip_data = ip_info.get_ip_info(arp_entry.ip, bypass_cache=bypass_cache)
        if include_hostname:
            arp_entry.hostname = ip_data.get('hostname')
        if include_mac_vendor:
            arp_entry.vendor = ip_data.get('vendor')
        updated_list.append(arp_entry)

    return updated_list

def _get_target_protocol_address_pdst() -> str:
    """
    _summary_

    Returns:
        str: _description_
    """
    ip = get_local_ip()
    octet = ip.split('.')
    octet[3] = '1'
    network = '.'.join(octet) + "/24"

    return network

def _ipv4_mask_len(dotquad) -> int:
    """
    Finds the number of bits set in the netmask.

    >>> ipv4_mask_len("255.255.255.0")
    24
    >>> ipv4_mask_len("0.0.0.0")
    0
    >>> ipv4_mask_len("255.255.255.255")
    32
    >>> ipv4_mask_len("127.0.0.0")
    Traceback (most recent call last):
    ...
    ValueError: Invalid netmask: 127.0.0.0
    """
    if not is_ipv4_address(dotquad):
        raise ValueError("Invalid netmask: {0}".format(dotquad))
    a, b, c, d = (int(octet) for octet in dotquad.split("."))
    mask = a << 24 | b << 16 | c << 8 | d

    if mask == 0:
            return 0

    # Count the number of consecutive 0 bits at the right.
    # https://wiki.python.org/moin/BitManipulation#lowestSet.28.29
    m = mask & -mask
    right0bits = -1
    while m:
        m >>= 1
        right0bits += 1

    # Verify that all the bits to the left are 1's
    if mask | ((1 << right0bits) - 1) != 0xffffffff:
        raise ValueError("Invalid netmask: {0}".format(dotquad))
    return 32 - right0bits

def _arp_entries_command() -> str:
    return "arp -a" if OSHelper().is_windows() else "/usr/sbin/arp -n"

def format_mac(mac: str) -> str:
    """
    Format mac str to OS style

    Args:
        mac (str): mac str representation (12 or 17 characters)

    Raises:
        ValueError: Invalid mac string

    Returns:
        str: formatted as XX:XX:XX:XX (linux) or XX-XX-XX-XX (win)
    """
    if len(mac) == 17:
        sep = mac[2]
        fmt_mac = mac.replace(sep, _mac_separator()).lower()
    elif len(mac) == 12:
        # sep = _mac_platform_separator()
        mac_byte_list = [mac[i:i+2] for i in range(0, 12, 2) ]
        fmt_mac = _mac_separator().join(mac_byte_list).lower()
    else:
        raise ValueError(f'MAC invalid format: {mac}')
    return fmt_mac

def unformat_mac(mac: str) -> str:
    """
    Unformat mac address by removeing seperators from string

    Args:
        mac (str): formatted mac str

    Returns:
        str: mac address with no seperators (i.e. ':' linux or '-' win)
    """
    mac_address = mac
    # Check macaddress format and try to compensate.
    if len(mac_address) == 12:
        LOGGER.debug('unformat_mac() - already unformattted')
    elif len(mac_address) == 12 + 5:
        mac_address = mac_address.replace(":", "").replace("-","")
            
    if len(mac_address) != 12:
        LOGGER.debug(f'unformat_mac(): unknown format - {mac} / {mac_address}')
        mac_address = mac

    LOGGER.debug(f'unformat_mac()  in: {mac}  out: {mac_address}')
    return mac_address

def _mac_separator() -> str:
    return '-' if OSHelper.is_windows() else ":"


if __name__ == "__main__":
    import dt_tools.cli.demos.dt_net_demos as cli
    import dt_tools.logger.logging_helper as lh
    lh.configure_logger()
    cli.demo()